"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs-extra");
const path = require("path");
const helpers_1 = require("../../helpers");
const logger_1 = require("../../logger");
function copyFile(dst, src) {
    logger_1.logger.action("copy", dst);
    fs.copyFileSync(src, dst);
}
function copyDir(dst, src) {
    logger_1.logger.action("copy", dst);
    fs.copySync(src, dst);
}
function copyFileIfExists(dst, src) {
    if (fs.existsSync(src)) {
        copyFile(dst, src);
    }
}
function mkdir(dir) {
    logger_1.logger.action("mkdir", dir);
    fs.mkdirpSync(dir);
}
function genFile(filepath, body) {
    logger_1.logger.action("create", filepath);
    fs.writeFileSync(filepath, body);
}
function genJsonFile(filepath, json) {
    genFile(filepath, JSON.stringify(json, null, 2));
}
const FUNCTIONS_PACKAGE_JSON = {
    private: true,
    name: "functions",
    engines: {
        node: "10",
    },
    dependencies: {
        "firebase-admin": "^8.0.0",
        "firebase-functions": "^3.0.0"
    }
};
const EXTRA_FIREBASE_FILES = [
    "firestore.rules",
    "firestore.indexes.json",
];
async function deploy(appDir, firmwarePath, opts) {
    if (!opts.firebaseProject) {
        throw new Error("--firebase-project is not set.");
    }
    logger_1.logger.progress("Packing the app...");
    const buildDir = await pack(appDir, firmwarePath, opts);
    // TODO: make sure that firebase-cli is installed.
    logger_1.logger.progress("Deploying...");
    helpers_1.exec(["firebase", "deploy"], { cwd: buildDir });
}
exports.deploy = deploy;
async function pack(appDir, firmwarePath, opts) {
    const buildDir = helpers_1.createTmpDir("makestack-firebase-pack");
    const appFilePath = (relpath) => path.join(appDir, relpath);
    const buildFilePath = (relpath) => path.join(buildDir, relpath);
    copyFile(buildFilePath("firebase.json"), appFilePath("firebase.json"));
    for (const filename of EXTRA_FIREBASE_FILES) {
        copyFileIfExists(buildFilePath(filename), appFilePath(filename));
    }
    mkdir(buildFilePath("public"));
    if (fs.existsSync(appFilePath("public"))) {
        for (const basename of fs.readdirSync(appFilePath("public"))) {
            const dst = buildFilePath(path.join("public", basename));
            copyFile(dst, appFilePath(path.join("public", basename)));
        }
    }
    const packageJson = FUNCTIONS_PACKAGE_JSON;
    const appDependencies = fs.readJsonSync(appFilePath("package.json")).dependencies;
    const makestackDpendencies = fs.readJsonSync(path.resolve(__dirname, "../../../package.json")).dependencies;
    packageJson.dependencies = Object.assign({}, makestackDpendencies, packageJson.dependencies, appDependencies);
    const indexJs = fs.readFileSync(path.join(__dirname, "start.js"), "utf-8");
    mkdir(buildFilePath("functions"));
    genJsonFile(buildFilePath("functions/package.json"), packageJson);
    genFile(buildFilePath("functions/index.js"), indexJs);
    copyFile(buildFilePath("functions/app.js"), appFilePath("app.js"));
    copyFile(buildFilePath("functions/firmware.bin"), firmwarePath);
    mkdir(buildFilePath("functions/makestack"));
    copyDir(buildFilePath("functions/makestack/dist"), path.resolve(__dirname, "../../../dist"));
    copyFile(buildFilePath("functions/makestack/package.json"), path.resolve(__dirname, "../../../package.json"));
    genJsonFile(buildFilePath(".firebaserc"), {
        projects: {
            default: opts.firebaseProject
        },
    });
    helpers_1.exec(["yarn", "install"], { cwd: buildFilePath("functions") });
    return buildDir;
}
function log(appDir, opts) {
    helpers_1.exec(["firebase", "functions:log", "--project", opts.firebaseProject]);
}
exports.log = log;
//# sourceMappingURL=firebase.js.map