"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const path = require("path");
class Command {
}
exports.Command = Command;
exports.validateAppDir = (appDir) => {
    if (!fs.existsSync(appDir)) {
        throw new Error("The app directory does not exist");
    }
    appDir = path.resolve(appDir);
    if (!fs.statSync(appDir).isDirectory()) {
        throw new Error(`${appDir} is not a directory`);
    }
    const packageJson = path.join(appDir, "package.json");
    if (!fs.existsSync(packageJson)) {
        throw new Error(`${packageJson} does not exist`);
    }
    const appJs = path.join(appDir, "app.js");
    const appTs = path.join(appDir, "app.ts");
    if (!fs.existsSync(appJs) && !fs.existsSync(appTs)) {
        throw new Error(`Neither ${appJs} nor ${appTs} does not exist.`);
    }
    return appDir;
};
const availableBoardTypes = ["esp32"];
exports.validateBoardType = (boardType) => {
    if (!availableBoardTypes.includes(boardType)) {
        throw new Error("Invalid board type.");
    }
    return require(`../boards/${boardType}`);
};
const availableCloudTypes = ["firebase"];
exports.validateCloudType = (cloudType) => {
    if (!availableCloudTypes.includes(cloudType)) {
        throw new Error("Invalid cloud type.");
    }
    return require(`../clouds/${cloudType}`);
};
const deviceFilePatterns = [
    /tty\.usbserial-.+/,
    /ttyUSB[0-9]/,
];
exports.validateDeviceFilePath = (deviceFile) => {
    if (deviceFile.length > 0) {
        return deviceFile;
    }
    const candidates = fs.readdirSync("/dev").filter((file) => {
        return deviceFilePatterns.some((pat) => pat.exec(file) !== null);
    });
    if (candidates.length === 0) {
        throw new Error("Failed to locate the device file.");
    }
    if (candidates.length > 1) {
        throw new Error("Found multiple device files. Please specify the one in the command-line option.");
    }
    return path.join("/dev", candidates[0]);
};
exports.APP_OPTS = [
    {
        name: "--app-dir <path>",
        desc: "The app directory.",
        default: process.cwd(),
        validator: exports.validateAppDir,
    },
];
exports.ADAPTER_OPTS = [
    {
        name: "--adapter <adapter>",
        desc: "The adapter type ('serial' or 'http').",
        default: "serial",
    },
];
exports.BOARD_OPTS = [
    // TODO: get the board type from package.json
    {
        name: "--board <board>",
        desc: "The board type (only 'esp32' for now).",
        default: "esp32",
        validator: exports.validateBoardType,
    },
];
exports.DEVICE_FILE_OPTS = [
    {
        name: "--device <path>",
        desc: "The device file path.",
        default: "",
        validator: exports.validateDeviceFilePath,
    },
    {
        name: "--baudrate <rate>",
        desc: "The baudrate.",
        default: 115200,
    },
];
exports.BUILD_OPTS = [
    {
        name: "--heartbeat-interval <seconds>",
        desc: "The interval of the heartbeat periodically sent from a device.",
        default: 30,
    },
    ...exports.APP_OPTS,
    ...exports.ADAPTER_OPTS,
    ...exports.BOARD_OPTS
];
exports.WATCH_OPTS = [
    {
        name: "--watch",
        desc: "Watch files for changes and ru-run the command.",
        default: false,
    }
];
exports.CLOUD_OPTS = [
    {
        name: "--cloud <cloud>",
        desc: "The cloud (only 'firebase' for now).",
        default: "firebase",
        validator: exports.validateCloudType,
    },
    {
        name: "--firebase-project <name>",
        desc: "The Firebase project name.",
    },
];
//# sourceMappingURL=command.js.map