"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const path = require("path");
const express = require("express");
const command_1 = require("./command");
const boards_1 = require("../boards");
const logger_1 = require("../logger");
const firmware_1 = require("../firmware");
const adapters_1 = require("../adapters");
const dev_server_1 = require("../dev_server");
const server_1 = require("../server/server");
class DevCommand extends command_1.Command {
    async run(_args, opts) {
        this.board = opts.board;
        // First, build the firmware. We need the firmware file in order to send
        // the latest version info in a heartbeat.
        if (!(await this.build(opts.appDir, opts))) {
            logger_1.logger.error("fix build errors and run the command again");
            process.exit(1);
        }
        const httpServerPort = opts.port + 1;
        const httpServer = express();
        new adapters_1.HTTPAdapter(httpServer, payload => {
            let reply = this.protocolServer.processPayload(payload);
            return reply ? reply : this.protocolServer.buildHeartbeatPayload();
        });
        httpServer.listen(httpServerPort, "127.0.0.1");
        logger_1.logger.progress("Initializing the adapter...");
        await this.initializeAdapter(opts.adapter, opts);
        logger_1.logger.progress(`Starting a app...`);
        this.devServer = new dev_server_1.DevServer(opts.host, opts.port, httpServerPort, opts.appDir);
        logger_1.logger.progress(`Listen on ${opts.host}:${opts.port}`);
        // Watch for the app source files.
        fs.watch(opts.appDir, async (_event, filename) => {
            const appFile = path.join(opts.appDir, "app.js");
            if (filename == "app.js" && fs.existsSync(appFile)) {
                logger_1.logger.progress("Change detected, restarting and rebuilding the app...");
                this.devServer.restart();
                await this.build(opts.appDir, opts);
            }
        });
        logger_1.logger.success(`We're ready! Watching for changes on ${opts.appDir}...`);
    }
    async build(appDir, opts) {
        logger_1.logger.progress("Building the firmware...");
        try {
            await firmware_1.buildApp(this.board, appDir, opts);
        }
        catch (e) {
            if (e instanceof boards_1.BuildError) {
                logger_1.logger.error("failed to build");
                return false;
            }
            else {
                throw e;
            }
        }
        this.protocolServer = new server_1.ProtocolServer(this.board.getFirmwarePath(), (name, value) => {
            this.devServer.sendRequest({ type: "event", name, value });
        });
        logger_1.logger.success("Build succeeded");
        return true;
    }
    async initializeAdapter(adapter, opts) {
        switch (adapter) {
            case "serial":
                // FIXME:
                const device = command_1.validateDeviceFilePath(opts.device || "");
                const serialAdapter = new adapters_1.SerialAdapter();
                await serialAdapter.open(device, opts.baudrate, payload => {
                    const reply = this.protocolServer.processPayload(payload);
                    if (reply) {
                        serialAdapter.send(reply);
                    }
                });
                // Send heartbeats regularly.
                // TODO: Disable heartbeaing on firmware updating.
                serialAdapter.send(this.protocolServer.buildHeartbeatPayload());
                setInterval(() => {
                    serialAdapter.send(this.protocolServer.buildHeartbeatPayload());
                }, 2000);
                // Make sure that connected device is running the our firmware.
                setTimeout(() => {
                    if (!this.protocolServer.verifiedPong) {
                        logger_1.logger.error("The device doesn't respond our health check.\nHint: Run `makestack flash --dev` to install the firmware.");
                        process.exit(1);
                    }
                }, 5000);
                break;
            case "http":
                /* Already running. */
                break;
            default:
                throw new Error(`Unknown adapter type: \`${adapter}'`);
        }
    }
}
DevCommand.command = "dev";
DevCommand.desc = "Starts up the app with auto update.";
DevCommand.args = [];
DevCommand.opts = [
    ...command_1.APP_OPTS,
    ...command_1.BOARD_OPTS,
    ...command_1.BUILD_OPTS,
    ...command_1.ADAPTER_OPTS,
    {
        name: "--device <path>",
        desc: "The device file path.",
    },
    {
        name: "--baudrate <rate>",
        desc: "The baudrate.",
        default: 115200,
    },
    {
        name: "--host <host>",
        desc: "The dev server hostname.",
        default: "0.0.0.0",
    },
    {
        name: "--port <port>",
        desc: "The dev server port.",
        default: 1234,
    },
];
DevCommand.watchMode = true;
exports.DevCommand = DevCommand;
//# sourceMappingURL=dev_command.js.map