"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const path = require("path");
const nunjucks = require("nunjucks");
const command_1 = require("./command");
const logger_1 = require("../logger");
const helpers_1 = require("../helpers");
const PACKAGE_JSON = `\
{
    "private": true,
    "makestack": {
        "name": "{{ name }}",
        "board": "esp32",
        "cloud": "firebase"
    },
    "scripts": {
{% if typescript %}
        "build": "./node_modules/.bin/tsc app.ts"
{% endif %}
    }
}
`;
const APP_JS = `\
const app = require("makestack")

app.onReady((device) => {
    const LED_PIN = 22
    device.pinMode(LED_PIN, "OUTPUT")
    while (1) {
        device.print("Blinking!")
        device.digitalWrite(LED_PIN, true)
        device.delay(1000)
        device.digitalWrite(LED_PIN, false)
        device.delay(1000)
    }
})
`;
const APP_TS = `\
import * as app from "makestack";
import { DeviceAPI } from "makestack";

app.onReady((device: DeviceAPI) => {
    const LED_PIN = 22;
    device.pinMode(LED_PIN, "OUTPUT");
    while (1) {
        device.print("Blinking!");
        device.digitalWrite(LED_PIN, true);
        device.delay(1000);
        device.digitalWrite(LED_PIN, false);
        device.delay(1000);
    }
});
`;
const TSCONFIG_JSON = `\
{
    "compilerOptions": {
        "declaration": true,
        "sourceMap": true,
        "module": "commonjs",
        "moduleResolution": "node",
        "pretty": true,
        "alwaysStrict": true,
        "strict": true,
        "target": "es2018"
    },
    "include": [
        "./src/**/*"
    ]
}
`;
const FIREBASE_JSON = `\
{
    "functions": {
    },
    "hosting": {
        "public": "public",
        "ignore": [
            "firebase.json",
            "**/.*",
            "**/node_modules/**"
        ]
    }
}
`;
const INDEX_HTML = `\
<!DOCTYPE html>
<html>
<head>
    <meta charset="utf-8">
    <title>{{ name }} - index.html</title>
    <style>
        body {
            font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, Oxygen, Ubuntu, Cantarell, 'Open Sans', 'Helvetica Neue', sans-serif;
            color: #121212;
        }
    </style>
</head>
<body>
    <div style="width: 500px; margin: 70px auto 0;">
        <h1 style="text-align: center;">{{name }} - index.html</h1>
        <hr>
        <p>
            The <code>makestack deploy</code> command will automatically uploads
            files in this <code>public</code> directory to the platform, for instance,
            <a href="https://firebase.google.com/docs/hosting">Firebase Hosting</a>.
        </p>
    </div>
</body>
</html>
`;
const GITIGNORE = `\
node_modules
*.log
app.js
`;
const DEPENDENCIES = ["makestack"];
const DEV_DEPENDENCIES = [];
const TYPESCRIPT_DEV_DEPENDENCIES = ["typescript"];
function genFile(filepath, template, ctx) {
    logger_1.logger.action("create", filepath);
    nunjucks.configure({ autoescape: false });
    fs.writeFileSync(filepath, nunjucks.renderString(template, ctx));
}
function mkdir(filepath) {
    logger_1.logger.action("mkdir", filepath);
    fs.mkdirSync(filepath);
}
function scaffold(appDir, opts) {
    if (fs.existsSync(appDir)) {
        throw new Error(`The directory alredy exists: \`${appDir}'`);
    }
    logger_1.logger.progress("Generating files");
    mkdir(appDir);
    const ctx = {
        name: path.basename(appDir),
        typescript: opts.typescript,
    };
    genFile(path.join(appDir, "package.json"), PACKAGE_JSON, ctx);
    genFile(path.join(appDir, ".gitignore"), GITIGNORE, ctx);
    if (opts.typescript) {
        genFile(path.join(appDir, "app.ts"), APP_TS, ctx);
        genFile(path.join(appDir, "tsconfig.json"), TSCONFIG_JSON, ctx);
    }
    else {
        genFile(path.join(appDir, "app.js"), APP_JS, ctx);
    }
    if (opts.firebase) {
        genFile(path.join(appDir, "firebase.json"), FIREBASE_JSON, ctx);
        mkdir(path.join(appDir, "public"));
        genFile(path.join(appDir, "public/index.html"), INDEX_HTML, ctx);
    }
    logger_1.logger.progress("Installing dependencies...");
    let deps = [...DEPENDENCIES, ...DEV_DEPENDENCIES];
    if (opts.typescript) {
        deps = deps.concat(TYPESCRIPT_DEV_DEPENDENCIES);
    }
    helpers_1.exec(["yarn", "add", ...deps], { cwd: appDir });
    logger_1.logger.success(`Successfully generated ${appDir}`);
}
class NewCommand extends command_1.Command {
    async run(args, opts) {
        await scaffold(args.path, opts);
    }
}
NewCommand.command = "new";
NewCommand.desc = "Create a new app.";
NewCommand.args = [
    {
        name: "path",
        desc: "The app directory.",
    }
];
NewCommand.opts = [
    {
        name: "--typescript",
        desc: "Create a TypeScript app.",
        default: false,
    },
    {
        name: "--firebase",
        desc: "Generate Firebase files.",
        default: false,
    },
];
exports.NewCommand = NewCommand;
//# sourceMappingURL=new_command.js.map