"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const zlib = require("zlib");
const protocol_1 = require("./protocol");
const helpers_1 = require("../helpers");
const logger_1 = require("../logger");
const firmware_1 = require("../firmware");
class ProtocolServer {
    constructor(firmwarePath, eventCallback) {
        this.verifiedPong = false;
        const firmwareVersion = firmware_1.extractCredentials(firmwarePath).version;
        const firmwareImage = fs.readFileSync(firmwarePath);
        this.firmwareVersion = firmwareVersion;
        this.firmwareImage = firmwareImage;
        this.eventCallback = eventCallback;
    }
    buildHeartbeatPayload() {
        return protocol_1.constructPayload({
            version: this.firmwareVersion,
            ping: {
                data: Buffer.from("HELO"),
            },
            corruptRateCheck: {
                length: 512,
            },
        });
    }
    processPayload(rawPayload) {
        const payload = protocol_1.parsePayload(rawPayload);
        if (payload.deviceStatus) {
            const { ramFree } = payload.deviceStatus;
            console.log(`ram free: ${ramFree} bytes (${helpers_1.bytesToReadableString(ramFree)})`);
        }
        if (payload.log) {
            for (const line of payload.log.split("\n")) {
                const EVENT_REGEX = /^@(?<name>[^ ]+) (?<type>[bis]):(?<value>.*)$/;
                const m = line.match(EVENT_REGEX);
                if (m) {
                    const { name, type, value: valueStr } = m.groups;
                    let value;
                    switch (type) {
                        case "b":
                            value = (valueStr == "true");
                            break;
                        case "i":
                            value = parseInt(valueStr);
                            break;
                        case "s":
                            value = valueStr;
                            break;
                        default:
                            logger_1.logger.warn(`unknown event type: \`${type}'`);
                            continue;
                    }
                    console.log(`event: name=${name}, value=${value}`);
                    this.eventCallback(name, value);
                }
                else {
                    console.log("device log:", line);
                }
            }
        }
        if (payload.pong) {
            this.verifiedPong = true;
        }
        if (payload.firmwareRequest) {
            if (payload.firmwareRequest.version != this.firmwareVersion) {
                logger_1.logger.warn(`invalid version: ${payload.firmwareRequest.version}`);
                return null;
            }
            const offset = payload.firmwareRequest.offset;
            const DATA_LEN = 8192;
            const data = this.firmwareImage.slice(offset, offset + DATA_LEN);
            let firmwareDataPayload;
            if (data.length > 0) {
                const compressed = zlib.deflateSync(data);
                if (compressed.length < data.length) {
                    firmwareDataPayload = protocol_1.constructPayload({
                        firmwareData: {
                            offset,
                            type: "deflate",
                            data: compressed,
                        },
                    });
                }
                else {
                    firmwareDataPayload = protocol_1.constructPayload({
                        firmwareData: { offset, type: "raw", data },
                    });
                }
            }
            else {
                firmwareDataPayload = protocol_1.constructPayload({
                    firmwareData: { offset, type: "eof", data },
                });
            }
            console.log(`Uploading len=${data.length}, offset=${offset} (${Math.floor((offset / this.firmwareImage.length) * 100)}%)`);
            return firmwareDataPayload;
        }
        return null;
    }
}
exports.ProtocolServer = ProtocolServer;
//# sourceMappingURL=server.js.map